/*******************************************************************
 *	Copyright (c) 1994-1998 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File:		driver/bc_dev.c
 *
 *	Description:	BestCrypt pseudo-device driver.
 *
 *	Scope:		BestCrypt pseudo-device driver
 *
 *	Platforms:	Linux
 * 
 *	Ideas:		Igor M. Arsenin
 *			Serge A. Frolov
 *			Vitaliy G. Zolotarev
 *
 *	Author:		Nail R. Kaipov
 *
 *	Created:	10-Nov-1998
 *
 *	Revision:	06-Aug-1999.	new strategy routine
 *			15-Sep-1999.	2.3.x kernels support
 *			11-Oct-1999.	strategy bugfix
 *			15-Dec-1999.	driver refinement
 *			21-Mar-2000.	bugfix for large blocks (v0.4b-2)
 *
 *******************************************************************/

#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB
#endif

#include "bc_cfg.h"
#include <linux/module.h>
#include <linux/fs.h>			/* inode et al            */
#include <linux/proc_fs.h>		/* inode et al            */
#include <linux/vmalloc.h>		/* vmalloc()              */
#include <linux/kernel.h>		/* printk()               */
#include <asm/uaccess.h>		/* put_user()             */
#include <asm/atomic.h>			/* atomic operations      */
#include <linux/hdreg.h>		/* hd_geometry            */
#include <linux/file.h>			/* inode et al            */

#include "bc_types.h"			/* bc types               */
#include "bc_ioctl.h"			/*           ioctl codes  */
#include "bc_mgr.h"
#include "bc_dev.h"			/* bc types        codes  */

#define DEVICE_OFF(dev)
#include <linux/blk.h>

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

static struct bc_device	bc_dev		[MAX_BC_DEVS];
static int		bc_sizes	[MAX_MINORS];
static int		bc_blksizes	[MAX_MINORS];
struct request		*BC_CURRENT = NULL;

#define BC_BUFFERS (NR_REQUEST/2)
#define BC_BUFFER_SIZE 4096

struct bc_buffer_desc {
	atomic_t	count;
	char		*ptr;
} bc_buffers[BC_BUFFERS];

static ssize_t bc_read      (struct file *,   char *, size_t, loff_t *);
static ssize_t bc_write     (struct file *,   const char *, size_t, loff_t *);
static int     bc_ioctl     (struct inode *,  struct file *, unsigned int, unsigned long);
static int     bc_open      (struct inode *,  struct file *);	
static int     bc_release   (struct inode *,  struct file *);
static int     dummy_ioctl  (struct inode *i, struct file *f, u_int cmd, u_long arg) { return ENODEV; }
static int     dummy_open   (struct inode *i, struct file *f) { return ENODEV; }
static int     dummy_release(struct inode *i, struct file *f) { return ENODEV; }
static int     *save_blksize;
static int     *save_blksize_size;
static int     max_loop;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,32))

#define BC_GET_REQUEST_FN(_MAJOR_) blk_dev[_MAJOR_].request_queue.request_fn
static void    dummy_request_fn  (request_queue_t *rq) { return; }
static void    (*save_request_fn)(request_queue_t *rq) = dummy_request_fn;
static struct  proc_dir_entry    *proc_bc;
int bc_proc_info(char *buf, char **start, off_t offset, int length)
{
	return proc_bc_info(buf, start, offset, length, 0);
}

#else

#define BC_GET_REQUEST_FN(_MAJOR_) blk_dev[_MAJOR_].request_fn
static void    dummy_request_fn(void) { return; }
static void    (*save_request_fn)(void) = dummy_request_fn;

static struct proc_dir_entry	proc_bc = {
	low_ino:  0,			/* dynamic inode */
	namelen:  6,			/* name length   */
	name:     "bcrypt",
	mode:     S_IFREG | S_IRUGO,
	nlink:    1,			/* # of links    */
	uid:      0, 
	gid:      0,
	size:     0,
	ops:      NULL,			/* i_ops, leaf   */
	get_info: proc_bc_info		/* get_info      */
};
#endif

static struct file_operations save_fops = {
	read   :block_read,
	write  :block_write,
	ioctl  :dummy_ioctl,
	open   :dummy_open,
	release:dummy_release,
	fsync  :block_fsync
};

static struct file_operations bc_fops = {
	read   :bc_read,
	write  :bc_write,
	ioctl  :bc_ioctl,
	open   :bc_open,
	release:bc_release,
	fsync  :block_fsync
};

EXPORT_SYMBOL(register_bc_algo);
EXPORT_SYMBOL(unregister_bc_algo);
MODULE_AUTHOR("Jetico, Inc.");

/*-----------------------------------------------------------*/
static inline int bc_minor(kdev_t dev)
{
	return ((MINOR(dev) >= BC_FIRST) && (MINOR(dev) <= BC_LAST));
}

static inline int bc_major(kdev_t dev)
{
	return (MAJOR_NR == MAJOR(dev));
}

static inline struct bc_device *get_bcdev(kdev_t dev)
{
	if (bc_minor(dev))
		return (&bc_dev[MINOR(dev)-BC_FIRST]);
	else
		return NULL;
}

/*-----------------------------------------------------------*/

static void grab_requests();
static void bc_request();

static void do_bc_request() 
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,32))
#define BC_NULL NULL
#else
#define BC_NULL
#endif
	if (NULL == CURRENT)
		return;
	grab_requests();
	save_request_fn(BC_NULL);
	bc_request(BC_NULL);
#undef BC_NULL
}


static void grab_requests()
{
	register struct request *prev, *tail, *curr;
	
	if (NULL == (curr = CURRENT))
		return;
	if (NULL != (tail = BC_CURRENT))
		while (tail->next)
			tail = tail->next;
	prev = NULL;
	while (curr) {
		if (bc_minor(curr->rq_dev)) {
			if (tail) tail->next = curr;
			else      BC_CURRENT = curr;
			if (prev) prev->next = curr->next;
			else      CURRENT    = curr->next; 
			tail = curr;
		} else {
			if (prev) prev = prev->next;
			else      prev = CURRENT; 
		}
		curr = curr->next;
	}
	if (tail) tail->next = NULL;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,32))
static void bc_request(request_queue_t *abcd)
#else
static void bc_request()
#endif
{
	struct bc_device	*bc;
	struct request		*req;
	struct file		*file;
	size_t			len, ret, size;
	loff_t			pos;
	char			*dst, *buf = NULL;
	u16			iv[4];
	mm_segment_t		fs;
	int			uptodate, i;

repeat:
	if (!BC_CURRENT) return;
	req = BC_CURRENT;
	BC_CURRENT = req->next;
	uptodate = 0;
	bc = get_bcdev(req->rq_dev);
	if (!bc || !bc->bc_dentry || !bc->bc_dentry->d_inode) {
		goto error_out;
	} else if (WRITE == req->cmd) {
		if (bc->bc_flags & BC_FLAGS_READ_ONLY) {
			goto error_out;
		}
	} else if (READ != req->cmd) {
		printk(KERN_ERR "bc: unknown command (%d).", req->cmd);
		goto error_out;
	}
	pos   = ((long long)req->sector) * 512 + bc->bc_offset;
	len   = req->current_nr_sectors << 9;
	dst   = req->buffer;
	file  = bc->bc_file;
	iv[0] = iv[1] = iv[2] = iv[3] = req->sector & 0xFFFF;

	if (READ == req->cmd) {
		fs = get_fs();
		set_fs(KERNEL_DS);
		spin_unlock_irq(&io_request_lock);

		ret = file->f_op->read(file, dst, len, &pos);
		spin_lock_irq(&io_request_lock);
		set_fs(fs);
		if (ret != len) {
			printk(KERN_ERR "bc: read error wanted %ld, read %ld\n", 
				(unsigned long)len, (unsigned long)ret);
			goto error_out;
		}
		bc->bc_alg->decrypt(bc->bc_key, (char *)iv, dst, dst, len);
	} else {
		for (i = 0; i < BC_BUFFERS; i++) {
			if (atomic_dec_and_test(&bc_buffers[i].count)) {
				buf = bc_buffers[i].ptr;
				break;
			} else
				atomic_inc(&bc_buffers[i].count);
		}
		if (BC_BUFFERS == i) {
			printk(KERN_ERR "bc: insufficient resources\n");
			goto error_out;
		}
		size = BC_BUFFER_SIZE;
		while (len > 0) {
			if (size > len)
				size = len;
			bc->bc_alg->encrypt(bc->bc_key, (char *)iv, dst, buf, size);
			fs = get_fs();
			set_fs(KERNEL_DS);
			spin_unlock_irq(&io_request_lock);
			ret = file->f_op->write(file, buf, size, &pos);
			spin_lock_irq(&io_request_lock);
			set_fs(fs);
			if (ret != size) {
				printk(KERN_ERR "bc: write error wanted %ld, wrote %ld\n", 
					(unsigned long)size, (unsigned long)ret);
				atomic_inc(&bc_buffers[i].count);
				goto error_out;
			}
			iv[0] = ((int)iv[0] + BC_BUFFER_SIZE / 512) & 0xFFFF;
			iv[3] = iv[2] = iv[1] = iv[0];
			dst += size;
			len -= size;
		}
		atomic_inc(&bc_buffers[i].count);
	}
	uptodate = 1;
error_out:	
        if (!end_that_request_first(req, uptodate, DEVICE_NAME))
		end_that_request_last(req);
	
	goto repeat;
}

/*-- end of strategy routine --------------------------------*/



/*-- fops functions -----------------------------------------*/
static int bc_open(struct inode *inode, struct file *file)
{
	struct bc_device *bc;

	if (!inode)	return EINVAL;
	if (!bc_major(inode->i_rdev)) {
		printk(KERN_ERR "bc_open: invalid major (%d).\n", MAJOR_NR);
		return ENODEV;
	}
	if (!bc_minor(inode->i_rdev))
		return save_fops.open(inode, file);
	else
		bc = get_bcdev(inode->i_rdev);
	bc->bc_refcnt++;
	MOD_INC_USE_COUNT;
	return 0;
}

static int bc_release(struct inode *inode, struct file *file)
{
	struct bc_device *bc;
	int error;
	
	if (!inode || !bc_major(inode->i_rdev)) 
		return 0;
	if (!bc_minor(inode->i_rdev))
		return save_fops.release(inode, file);
	else
		bc = get_bcdev(inode->i_rdev);

	error = fsync_dev(inode->i_rdev);

	if (bc->bc_refcnt <= 0)	
		printk(KERN_WARNING "bc_release: refcount (%d) <= 0\n", bc->bc_refcnt);
	else  {
		bc->bc_refcnt--;
		MOD_DEC_USE_COUNT;
	}
	return error;
}

static ssize_t bc_read (struct file *file, char *buf, size_t count, loff_t *ppos)
{
	struct bc_device *bc;
	struct inode	*inode;
	
	if (!file)	
		return EINVAL;
	
	inode = file->f_dentry->d_inode;	

	if (!bc_major(inode->i_rdev))
		return ENODEV;

	if (NULL == (bc = get_bcdev(inode->i_rdev)))
		return save_fops.read(file, buf, count, ppos);

	if (bc->bc_flags & BC_FLAGS_DEV_LOCK)
		return EIO;

	return block_read(file, buf, count, ppos);
}

static ssize_t bc_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	struct bc_device *bc;
	struct inode	*inode;
		
	if (!file)
		return EINVAL;
		
	inode = file->f_dentry->d_inode;
	
	if (!bc_major(inode->i_rdev))
		return ENODEV;
	if (NULL == (bc = get_bcdev(inode->i_rdev)))
		return save_fops.write(file, buf, count, ppos);

	if (bc->bc_flags & BC_FLAGS_DEV_LOCK)
		return EIO;

	return block_write(file, buf, count, ppos);
}

/*-- end of fops functions ----------------------------------*/

#define BC_HANDLER(x, y)        case (x): return (y)
#define BC_GET_ARG(arg, query)  if (NULL == (arg)) return EINVAL; \
                                if (copy_from_user(&(query), arg, sizeof(query))) \
					return EFAULT;
/*-----------------------------------------------------------*/

static int bc_get_info(struct bc_device *bc, struct bc_info *arg)
{
	struct bc_info	query;
	
	BC_GET_ARG(arg, query)

	put_user(BC_VERSION_MAJOR, (long *)&arg->ver_major);
	put_user(BC_VERSION_MINOR, (long *)&arg->ver_minor);
	put_user(bc->bc_flags & BC_FLAGS_DEV_BUSY ? 1 : 0,  (long *)&arg->busy);

	return 0;
}

static int bc_check(struct bc_device *bc, struct bc_check *arg)
{
	struct bc_check	query;
	struct file	*file;
	int             i;
	
	BC_GET_ARG(arg, query)

	if (!(file = fget(query.fd)))
		return EBADF;

	for (i = 0; i < MAX_BC_DEVS; i++) {
		if (NULL == bc_dev[i].bc_dentry)
			continue;
		if (file == bc_dev[i].bc_file)
			break;
	}
	if (MAX_BC_DEVS == i)
		i = BC_CONTAINER_UNUSED;
	else if (bc_dev[i].bc_flags & BC_FLAGS_DEV_LOCK)
		i = BC_CONTAINER_MOUNTED;
	else
		i = BC_CONTAINER_USED;

	put_user(i, (long *)&arg->busy);

	return 0;
}


#define MAX_DISK_SIZE 1024*1024*1024

static  void figure_bc_size(struct bc_device *bc)
{
	int	size;
				
	if (S_ISREG(bc->bc_dentry->d_inode->i_mode)) {
		size = (bc->bc_dentry->d_inode->i_size - bc->bc_offset) >> BLOCK_SIZE_BITS;
	} else {	/* special inode */
		kdev_t bcdev = bc->bc_dev;
		if (blk_size[MAJOR(bcdev)])
			size = blk_size[MAJOR(bcdev)][MINOR(bcdev)] - (bc->bc_offset >> BLOCK_SIZE_BITS);
		else
			size = MAX_DISK_SIZE;
	}
	bc_sizes[bc->bc_number] = size;
}


static int bc_set_fd(struct bc_device *bc, kdev_t dev, struct bc_file *arg)
/* key must be ready */
{
	struct file	*file;
	struct inode	*inode;
	struct bc_algorithm	*bc_alg;
	struct bc_file	query; 
	int error;

	BC_GET_ARG(arg, query)

	if (bc->bc_dentry)
		return EBUSY;
		
	bc_alg = get_bc_algo(query.alg_id, NULL);
	if (!bc_alg || bc_alg->test_key(query.key_handle) != 0) 
		return EINVAL;

	if (!(file = fget(query.fd)))
		return EBADF;
	inode = file->f_dentry->d_inode;
	error = EINVAL;
	if (!inode) {
		printk(KERN_ERR "bc_set_fd: NULL inode.\n");
		goto error_out;
	}
	if (!file->f_op) {
		printk(KERN_ERR "bc_set_fd: NULL file_operations.\n");
		goto error_out;
	}
	if (!file->f_op->read) {
		printk(KERN_ERR "bc_set_fd: Can't perform reads.\n");
		goto error_out;
	}

	if (S_ISBLK(inode->i_mode)) {
		error = blkdev_open(inode, file);
		if (error) 
			goto error_out;
		bc->bc_dev   = inode->i_rdev;
	} else if (S_ISREG(inode->i_mode)) {
		bc->bc_dev   = inode->i_dev;
	} else 
		goto error_out;

	if (IS_RDONLY (inode) || is_read_only(bc->bc_dev) || (NULL == file->f_op->write)) {
		bc->bc_flags = BC_FLAGS_READ_ONLY;
		set_device_ro(dev, 1);
	} else {
		bc->bc_flags = 0;
		invalidate_inode_pages (inode);
		set_device_ro(dev, 0);
	}
	
	bc->bc_flags	|= query.flags & BC_USER_FLAGS_MASK;
	bc->bc_flags	|= BC_FLAGS_DEV_BUSY;
	bc->bc_offset	= query.offset;
	bc->bc_key	= query.key_handle;
	bc->bc_self_dev	= dev;
	bc->bc_alg	= bc_alg;
	bc->bc_dentry	= dget(file->f_dentry);
	bc->bc_file	= file; 

	figure_bc_size(bc);
	bc_alg->lock_key(query.key_handle, 1);
	MOD_INC_USE_COUNT;

	return 0;

error_out:
	if (file)
		fput(file);
	return error;
}

static int bc_clr_fd(struct bc_device *bc, kdev_t dev)
{
	if (!bc->bc_dentry)
		return ENXIO;
	if (bc->bc_refcnt > 1)	/* we needed one fd for the ioctl */
		return EBUSY;

	if (S_ISBLK(bc->bc_dentry->d_inode->i_mode))	
		blkdev_release (bc->bc_dentry->d_inode);
	dput(bc->bc_dentry);	/* release inode */
	fput(bc->bc_file);
	
	bc->bc_flags    = 0;
	bc->bc_dev      = 0;
	bc->bc_self_dev = 0;
	bc->bc_dentry   = NULL;
	bc->bc_file     = NULL;
	bc->bc_offset   = 0;
	bc_sizes[bc->bc_number] = 0;
	invalidate_buffers(dev);
	bc->bc_alg->lock_key(bc->bc_key, 0);
	bc->bc_alg->free_key(bc->bc_key);
	bc->bc_alg = NULL;
	bc->bc_key = 0;
	MOD_DEC_USE_COUNT;

	return 0;
}

static int bc_lock_dev(struct bc_device * bc, int lock)
{
	struct super_block * sb;

	if (!capable(CAP_SYS_ADMIN))
		return EPERM;

	if (!(bc->bc_flags & BC_FLAGS_DEV_BUSY)) {
		if (!lock) 
			return 0;
		printk(KERN_ERR "bc: attempt to lock free device.\n");
		return EINVAL;
	}
	sb = get_super(bc->bc_self_dev);
	if (!sb || !sb->s_root) {
		if (!lock)
			return 0;
		printk(KERN_ERR "bc: attempt to lock unmounted device.\n");
		return EINVAL;
	}

	if (lock && !(bc->bc_flags & BC_FLAGS_DEV_LOCK)) {
	    sb->s_root->d_count++;
	    bc->bc_flags |= BC_FLAGS_DEV_LOCK;
	} else if (!lock && bc->bc_flags & BC_FLAGS_DEV_LOCK) {
	    sb->s_root->d_count--;
	    bc->bc_flags &= ~BC_FLAGS_DEV_LOCK;
	}
	return 0;
}

/*-- algorithm driver ioctls --------------------------------*/

static int bc_verify_alg(struct bc_alg *arg)
{
	struct bc_alg		query;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC || 
	   !get_bc_algo(query.alg_id, &query.alg_module[0])) 
		return EINVAL;
	return 0;	
}

static int bc_make_key(struct bc_key *arg)
{
	struct bc_key		query;
	struct bc_algorithm	*bc_alg;
	int			error;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC || !query.key || !query.pool) 
		return EINVAL;

	bc_alg = get_bc_algo(query.alg_id, &query.alg_module[0]);
	if (!bc_alg) 
		return EINVAL;

	/* WARNING!!! using user mode memory areas */
	error = bc_alg->make_key(query.key, query.key_len, query.pool, &query.key_handle);
	if (error)
		return error;	

	if (copy_to_user(&arg->key_handle, &query.key_handle, sizeof(KEY_HANDLE)))
		return EFAULT;

	return 0;	
}

static int bc_free_key(struct bc_key *arg)
{
	struct bc_key		query;
	struct bc_algorithm	*bc_alg;
	int			error;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC) 
		return EINVAL;

	bc_alg = get_bc_algo(query.alg_id, NULL);
	if (!bc_alg) 
		return EINVAL;
	error = bc_alg->free_key(query.key_handle);
	return error;
}

static int bc_process(struct bc_block *arg, int operation)
{
	struct bc_block		query;
	struct bc_algorithm	*bc_alg;
	int			error;
	char			*buffer;

	BC_GET_ARG(arg, query)

	if (query.magic != BC_MAGIC || 
	   !query.buffer || 
	   query.buffer_len > MAX_BLOCK_SIZE) 
		return EINVAL;
	
	bc_alg = get_bc_algo(query.alg_id, NULL);
	if (!bc_alg) 
		return EINVAL;

	buffer = vmalloc(query.buffer_len);
	if (!buffer) 
		return ENOMEM;
	
	if (copy_from_user(buffer, query.buffer, query.buffer_len)) {
		error = EFAULT;
		goto error_out;
	}

	switch (operation) {
	case BC_ENCRYPT_BLOCK:
		bc_alg->encrypt(query.key_handle, &query.iv[0], 
				buffer, buffer, query.buffer_len);
		break;
	case BC_DECRYPT_BLOCK:
		bc_alg->decrypt(query.key_handle, &query.iv[0], 
				buffer, buffer, query.buffer_len);
		break;
	default:
		error = EINVAL;
		goto error_out;
	}
	
	error = copy_to_user(query.buffer, buffer, query.buffer_len);

error_out:
	if (buffer) {
		memset(buffer, 0, query.buffer_len);
		vfree(buffer);
		buffer = NULL;
	}
	return error;
}
/*--------------------------------------------------------------------*/
static int blkgetsize(struct bc_device *bc, long *arg) 
{
	if (!bc->bc_dentry)	
		return ENXIO;
	if (!arg)
		return EINVAL;
	put_user(bc_sizes[bc->bc_number] << 1, arg);
	return 0;
}

static int hdio_getgeo(struct bc_device *bc, struct hd_geometry *geo) 
{
	unsigned long tmp;

	if (!bc->bc_dentry)
		return ENXIO;
	if (!geo)
		return EINVAL;
	tmp = 0x80*0x80*0x8000 - bc_sizes[bc->bc_number];
	put_user (0x80, (u8 *)&geo->heads);
	put_user (0x80, (u8 *)&geo->sectors);
	put_user (0x8000, (u16 *)&geo->cylinders);
	put_user (tmp, (u32 *)&geo->start);
	return 0;
}
/*====================================================================*/
static int bc_ioctl(struct inode *inode, struct file *file, u_int cmd, u_long arg)
{
	struct bc_device *bc;
	int err;
	
	if (!inode) 
		return EINVAL;
	if (!bc_major(inode->i_rdev))
		return ENODEV;
		
        if (!bc_minor(inode->i_rdev)) {
		err = save_fops.ioctl(inode, file, cmd, arg);
		if (max_loop) {
			if (save_blksize)
				memcpy(bc_sizes, save_blksize, (max_loop)*sizeof(int));
			if (save_blksize_size)
				memcpy(bc_blksizes, save_blksize_size, (max_loop)*sizeof(int));
		}
		return err;
	} else
		bc = get_bcdev(inode->i_rdev);

	switch (cmd) {
	BC_HANDLER(BC_GET_INFO,      bc_get_info(bc, (struct bc_info *) arg));
	BC_HANDLER(BC_SET_FD,        bc_set_fd(bc, inode->i_rdev, (struct bc_file *) arg));
	BC_HANDLER(BC_CLR_FD,        bc_clr_fd(bc, inode->i_rdev));
	BC_HANDLER(BC_LOCK_DEV,      bc_lock_dev(bc, 1));
	BC_HANDLER(BC_UNLOCK_DEV,    bc_lock_dev(bc, 0));
	BC_HANDLER(BC_CHECK_FD,      bc_check(bc, (struct bc_check *) arg));
	BC_HANDLER(BC_VERIFY_ALG,    bc_verify_alg((struct bc_alg*) arg));
	BC_HANDLER(BC_MAKE_KEY,      bc_make_key((struct bc_key*) arg));
	BC_HANDLER(BC_FREE_KEY,      bc_free_key((struct bc_key*) arg));
	BC_HANDLER(BC_ENCRYPT_BLOCK, bc_process((struct bc_block*) arg, BC_ENCRYPT_BLOCK));
	BC_HANDLER(BC_DECRYPT_BLOCK, bc_process((struct bc_block*) arg, BC_DECRYPT_BLOCK));
	BC_HANDLER(BLKGETSIZE,       blkgetsize(bc, (long *)arg));			/* Return device size */
	BC_HANDLER(HDIO_GETGEO,      hdio_getgeo(bc, (struct hd_geometry *) arg));	/* return pseudo geometry for mkfs.msdos */
	default:
		return EINVAL;
	}
	return 0;
}
/*====================================================================*/

/*-----------------------------------------------------------*/
int init_module(void) 
{
	int	rc;
	struct file_operations *fops;
	struct inode		fake_inode;

	printk(KERN_INFO "Loading %s ...\n", BC_VERSION_STRING);
	memset(&fake_inode, 0, sizeof(struct inode));
	fake_inode.i_rdev = MKDEV(MAJOR_NR, BC_FIRST);
	fops = get_blkfops(MAJOR_NR);
	if (NULL == fops) {
		if (0 > register_blkdev(MAJOR_NR, DEVICE_NAME, &bc_fops)) {
			printk(KERN_ERR "bc: unable to get major number %d\n", MAJOR_NR);
			return EIO;
		}
		
	} else {
		if (-ENODEV != fops->open(&fake_inode, NULL)) {
			fops->release(&fake_inode, NULL);
			printk(KERN_ERR "bc: already loaded\n");
			return EIO;
		}
		for (max_loop = 255; max_loop >= 0; max_loop--) {
			fake_inode.i_rdev = MKDEV(MAJOR_NR, max_loop);
			if (-ENODEV != fops->open(&fake_inode, NULL))
				break;
		}
		if ((max_loop < 0) || (max_loop >= BC_FIRST)) {
			printk(KERN_ERR "bc: initialization error\n");
			return EIO;
		}
		max_loop++;
		
		memcpy(&save_fops, fops, sizeof(struct file_operations));
		memcpy(fops, &bc_fops, sizeof(struct file_operations));
		save_request_fn = BC_GET_REQUEST_FN(MAJOR_NR);
 
		save_blksize = blk_size[MAJOR_NR];
		save_blksize_size = blksize_size[MAJOR_NR];
	}
	
	for (rc = 0; rc < MAX_BC_DEVS; rc++) {
		memset(&bc_dev[rc], 0, sizeof(struct bc_device));
		bc_dev[rc].bc_number = rc+BC_FIRST;
	}
	memset(&bc_sizes,    0, sizeof(bc_sizes));
	memset(&bc_blksizes, 0, sizeof(bc_blksizes));

	if (0 > init_bc_algo()) {
		rc = EINVAL; 
		goto error_out;
	}
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,32))
	proc_bc = create_proc_entry("bcrypt", S_IFREG | S_IRUGO, &proc_root);
	proc_bc->get_info = bc_proc_info;
#else
	proc_register(&proc_root, &proc_bc);
#endif

	if (max_loop) {
		if (blk_size[MAJOR_NR])
			memcpy(bc_sizes, blk_size[MAJOR_NR], (max_loop)*sizeof(int));
		if (blksize_size[MAJOR_NR])
			memcpy(bc_blksizes, blksize_size[MAJOR_NR], (max_loop)*sizeof(int));
	}
	BC_GET_REQUEST_FN(MAJOR_NR) = do_bc_request;
	blk_size[MAJOR_NR] = bc_sizes;
	blksize_size[MAJOR_NR] = bc_blksizes;
	
	bc_buffers[0].ptr = vmalloc(BC_BUFFERS * BC_BUFFER_SIZE);
	if (NULL == bc_buffers[0].ptr)
		goto error_out;
	/* init bc_buffers */
	for (rc = 0; rc < BC_BUFFERS; rc++) {
		bc_buffers[rc].ptr = bc_buffers[0].ptr + rc * BC_BUFFER_SIZE;
		atomic_set(&bc_buffers[rc].count, 1);
	}
	return 0;
	
error_out:
	if (NULL == fops) {
		unregister_blkdev(MAJOR_NR, DEVICE_NAME);	
	} else {
		memcpy(fops, &save_fops, sizeof(struct file_operations));
	}
	return rc;
}

void cleanup_module(void) 
{
	struct file_operations *fops;
	struct inode fake_inode;

	if (max_loop) {
		fops = get_blkfops(MAJOR_NR);
		memcpy(fops, &save_fops, sizeof(struct file_operations));
		BC_GET_REQUEST_FN(MAJOR_NR) = save_request_fn;
		memcpy(save_blksize, blk_size[MAJOR_NR], max_loop*sizeof(int));
		memcpy(save_blksize_size, blksize_size[MAJOR_NR], max_loop*sizeof(int));
		blk_size[MAJOR_NR] = save_blksize;
		blksize_size[MAJOR_NR] = save_blksize_size;
	
		memset(&fake_inode, 0, sizeof(struct inode));
		fake_inode.i_rdev = MKDEV(MAJOR_NR, max_loop-1);
		fops->release(&fake_inode, NULL);	
	} else {
		if (unregister_blkdev(MAJOR_NR, DEVICE_NAME) != 0) 
			printk(KERN_ERR "bc: cleanup failed.\n");
	}
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,32))
	remove_proc_entry("bcrypt", &proc_root);
#else
	proc_unregister(&proc_root, proc_bc.low_ino);
#endif
	if (NULL != bc_buffers[0].ptr) 
		vfree(bc_buffers[0].ptr);
}
/*-- end of device registration stuff -----------------------*/
